#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from unittest.mock import MagicMock

import pytest
import responses
from airbyte_cdk.utils import AirbyteTracedException
from source_pinterest.source import SourcePinterest


def setup_responses():
    responses.add(
        responses.POST,
        "https://api.pinterest.com/v5/oauth/token",
        json={"access_token": "fake_access_token", "expires_in": 3600},
    )
    responses.add(
        responses.GET,
        "https://api.pinterest.com/v5/user_account",
        json={},
    )


@responses.activate
def test_check_connection(test_config):
    setup_responses()
    source = SourcePinterest()
    logger_mock = MagicMock()
    assert source.check_connection(logger_mock, test_config) == (True, None)


def test_check_wrong_date_connection(wrong_date_config):
    source = SourcePinterest()
    logger_mock = MagicMock()
    with pytest.raises(AirbyteTracedException) as e:
        source.check_connection(logger_mock, wrong_date_config)
    assert e.value.message == "Entered `Start Date` wrong_date_format does not match format YYYY-MM-DD"


@responses.activate
def test_check_connection_expired_token(test_config):
    responses.add(responses.POST, "https://api.pinterest.com/v5/oauth/token", status=401)
    source = SourcePinterest()
    logger_mock = MagicMock()
    assert source.check_connection(logger_mock, test_config) == (
        False,
        "Try to re-authenticate because current refresh token is not valid. "
        "401 Client Error: Unauthorized for url: https://api.pinterest.com/v5/oauth/token",
    )


def test_get_authenticator(test_config):
    source = SourcePinterest()
    auth = source.get_authenticator(test_config)
    expected = test_config.get("refresh_token")
    assert auth.refresh_token == expected



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from http import HTTPStatus
from unittest.mock import MagicMock

import pytest
from airbyte_cdk.models import SyncMode
from pytest import fixture
from source_pinterest.streams import AdAccountAnalytics, Campaigns, IncrementalPinterestSubStream


@fixture
def patch_incremental_base_class(mocker):
    # Mock abstract methods to enable instantiating abstract class
    mocker.patch.object(IncrementalPinterestSubStream, "path", "v0/example_endpoint")
    mocker.patch.object(IncrementalPinterestSubStream, "primary_key", "test_primary_key")
    mocker.patch.object(IncrementalPinterestSubStream, "__abstractmethods__", set())


def test_cursor_field(patch_incremental_base_class):
    stream = IncrementalPinterestSubStream(None, config=MagicMock())
    expected_cursor_field = "updated_time"
    assert stream.cursor_field == expected_cursor_field


def test_get_updated_state(patch_incremental_base_class, test_current_stream_state):
    stream = IncrementalPinterestSubStream(None, config=MagicMock())
    inputs = {"current_stream_state": test_current_stream_state, "latest_record": test_current_stream_state}
    expected_state = {"updated_time": "2021-10-22"}
    assert stream.get_updated_state(**inputs) == expected_state


def test_stream_slices(patch_incremental_base_class, test_current_stream_state, test_incremental_config):
    stream = IncrementalPinterestSubStream(None, config=test_incremental_config)
    inputs = {"sync_mode": SyncMode.incremental, "cursor_field": "updated_time", "stream_state": test_current_stream_state}
    expected_stream_slice = {"start_date": "2021-10-22", "end_date": "2021-11-21"}
    assert next(stream.stream_slices(**inputs)) == expected_stream_slice


def test_supports_incremental(patch_incremental_base_class, mocker):
    mocker.patch.object(IncrementalPinterestSubStream, "cursor_field", "dummy_field")
    stream = IncrementalPinterestSubStream(None, config=MagicMock())
    assert stream.supports_incremental


def test_source_defined_cursor(patch_incremental_base_class):
    stream = IncrementalPinterestSubStream(None, config=MagicMock())
    assert stream.source_defined_cursor


def test_stream_checkpoint_interval(patch_incremental_base_class):
    stream = IncrementalPinterestSubStream(None, config=MagicMock())
    expected_checkpoint_interval = None
    assert stream.state_checkpoint_interval == expected_checkpoint_interval


def test_request_params(patch_incremental_base_class):
    stream = AdAccountAnalytics(None, config=MagicMock())
    test_slice = {"start_date": "2022-01-01", "end_date": "2022-01-02"}
    expected_property = "columns"
    res = stream.request_params({}, test_slice)
    assert expected_property in res


@pytest.mark.parametrize(
    ("http_status", "should_retry"),
    [
        (HTTPStatus.OK, False),
        (HTTPStatus.BAD_REQUEST, False),
        (HTTPStatus.TOO_MANY_REQUESTS, False),
        (HTTPStatus.INTERNAL_SERVER_ERROR, True),
    ],
)
def test_should_retry(patch_incremental_base_class, http_status, should_retry):
    response_mock = MagicMock()
    response_mock.status_code = http_status
    stream = AdAccountAnalytics(None, config=MagicMock())
    assert stream.should_retry(response_mock) == should_retry


def test_parse_response(patch_incremental_base_class, test_response_filter, test_current_stream_state):
    stream = Campaigns(None, config=MagicMock())
    expected_parsed_object = [{"updated_time": "2021-11-01"}]
    result = list(stream.parse_response(test_response_filter, test_current_stream_state))
    assert result == expected_parsed_object



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import copy
import logging
from base64 import standard_b64encode
from typing import Any, List, Mapping, Tuple, Type

import pendulum
import requests
from airbyte_cdk.models import FailureType
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.http.auth import Oauth2Authenticator
from airbyte_cdk.utils import AirbyteTracedException
from source_pinterest.reports import CampaignAnalyticsReport

from .reports.reports import (
    AdGroupReport,
    AdGroupTargetingReport,
    AdvertiserReport,
    AdvertiserTargetingReport,
    CampaignTargetingReport,
    CustomReport,
    KeywordReport,
    PinPromotionReport,
    PinPromotionTargetingReport,
    ProductGroupReport,
    ProductGroupTargetingReport,
    ProductItemReport,
)
from .streams import (
    AdAccountAnalytics,
    AdAccounts,
    AdAnalytics,
    AdGroupAnalytics,
    AdGroups,
    Ads,
    Audiences,
    BoardPins,
    Boards,
    BoardSectionPins,
    BoardSections,
    CampaignAnalytics,
    Campaigns,
    Catalogs,
    CatalogsFeeds,
    CatalogsProductGroups,
    ConversionTags,
    CustomerLists,
    Keywords,
    PinterestStream,
    UserAccountAnalytics,
)

logger = logging.getLogger("airbyte")


class SourcePinterest(AbstractSource):
    def _validate_and_transform(self, config: Mapping[str, Any], amount_of_days_allowed_for_lookup: int = 89):
        config = copy.deepcopy(config)
        today = pendulum.today()
        latest_date_allowed_by_api = today.subtract(days=amount_of_days_allowed_for_lookup)

        start_date = config.get("start_date")

        # transform to datetime
        if start_date and isinstance(start_date, str):
            try:
                config["start_date"] = pendulum.from_format(start_date, "YYYY-MM-DD")
            except ValueError:
                message = f"Entered `Start Date` {start_date} does not match format YYYY-MM-DD"
                raise AirbyteTracedException(
                    message=message,
                    internal_message=message,
                    failure_type=FailureType.config_error,
                )

        if not start_date or config["start_date"] < latest_date_allowed_by_api:
            logger.info(
                f"Current start_date: {start_date} does not meet API report requirements. Resetting start_date to: {latest_date_allowed_by_api}"
            )
            config["start_date"] = latest_date_allowed_by_api

        return config

    @staticmethod
    def get_authenticator(config):
        config = config.get("credentials") or config
        credentials_base64_encoded = standard_b64encode(
            (config.get("client_id") + ":" + config.get("client_secret")).encode("ascii")
        ).decode("ascii")
        auth = f"Basic {credentials_base64_encoded}"

        return Oauth2Authenticator(
            token_refresh_endpoint=f"{PinterestStream.url_base}oauth/token",
            client_secret=config.get("client_secret"),
            client_id=config.get("client_id"),
            refresh_access_token_headers={"Authorization": auth},
            refresh_token=config.get("refresh_token"),
        )

    def check_connection(self, logger, config) -> Tuple[bool, any]:
        config = self._validate_and_transform(config)
        authenticator = self.get_authenticator(config)
        url = f"{PinterestStream.url_base}user_account"
        try:
            auth_headers = {"Accept": "application/json", **authenticator.get_auth_header()}
            session = requests.get(url, headers=auth_headers)
            session.raise_for_status()
            return True, None
        except requests.exceptions.HTTPError as e:
            if "401 Client Error: Unauthorized for url" in str(e):
                return False, f"Try to re-authenticate because current refresh token is not valid. {e}"
            else:
                return False, e
        except requests.exceptions.RequestException as e:
            return False, e

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        config["authenticator"] = self.get_authenticator(config)
        report_config = self._validate_and_transform(config, amount_of_days_allowed_for_lookup=913)
        config = self._validate_and_transform(config)
        status = ",".join(config.get("status")) if config.get("status") else None

        ad_accounts = AdAccounts(config)
        ads = Ads(ad_accounts, config=config, status_filter=status)
        ad_groups = AdGroups(ad_accounts, config=config, status_filter=status)
        campaigns = Campaigns(ad_accounts, config=config, status_filter=status)
        boards = Boards(config)
        board_sections = BoardSections(boards, config=config)
        return [
            ad_accounts,
            AdAccountAnalytics(ad_accounts, config=config),
            ads,
            AdAnalytics(ads, config=config),
            ad_groups,
            AdGroupAnalytics(ad_groups, config=config),
            boards,
            BoardPins(boards, config=config),
            board_sections,
            BoardSectionPins(board_sections, config=config),
            campaigns,
            CampaignAnalytics(campaigns, config=config),
            CampaignAnalyticsReport(ad_accounts, config=report_config),
            CampaignTargetingReport(ad_accounts, config=report_config),
            UserAccountAnalytics(None, config=config),
            Keywords(ad_groups, config=config),
            Audiences(ad_accounts, config=config),
            ConversionTags(ad_accounts, config=config),
            CustomerLists(ad_accounts, config=config),
            Catalogs(config=config),
            CatalogsFeeds(config=config),
            CatalogsProductGroups(config=config),
            AdvertiserReport(ad_accounts, config=report_config),
            AdvertiserTargetingReport(ad_accounts, config=report_config),
            AdGroupReport(ad_accounts, config=report_config),
            AdGroupTargetingReport(ad_accounts, config=report_config),
            PinPromotionReport(ad_accounts, config=report_config),
            PinPromotionTargetingReport(ad_accounts, config=report_config),
            ProductGroupReport(ad_accounts, config=report_config),
            ProductGroupTargetingReport(ad_accounts, config=report_config),
            KeywordReport(ad_accounts, config=report_config),
            ProductItemReport(ad_accounts, config=report_config),
        ] + self.get_custom_report_streams(ad_accounts, config=report_config)

    def get_custom_report_streams(self, parent, config: dict) -> List[Type[Stream]]:
        """return custom report streams"""
        custom_streams = []
        for report_config in config.get("custom_reports", []):
            report_config["authenticator"] = config["authenticator"]

            # https://developers.pinterest.com/docs/api/v5/#operation/analytics/get_report
            if report_config.get("granularity") == "HOUR":
                # Otherwise: Response Code: 400 {"code":1,"message":"HOURLY request must be less than 3 days"}
                amount_of_days_allowed_for_lookup = 2
            elif report_config.get("level") == "PRODUCT_ITEM":
                amount_of_days_allowed_for_lookup = 91
            else:
                amount_of_days_allowed_for_lookup = 913

            start_date = report_config.get("start_date")
            if not start_date:
                report_config["start_date"] = config.get("start_date")

            report_config = self._validate_and_transform(report_config, amount_of_days_allowed_for_lookup)

            stream = CustomReport(
                parent=parent,
                config=report_config,
            )
            custom_streams.append(stream)
        return custom_streams



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from unittest.mock import MagicMock

from pytest import fixture
from source_pinterest.reports import CampaignAnalyticsReport


@fixture
def test_config():
    return {
        "client_id": "test_client_id",
        "client_secret": "test_client_secret",
        "refresh_token": "test_refresh_token",
        "start_date": "2021-05-07",
    }


@fixture
def wrong_date_config():
    return {
        "client_id": "test_client_id",
        "client_secret": "test_client_secret",
        "refresh_token": "test_refresh_token",
        "start_date": "wrong_date_format",
    }


@fixture
def test_incremental_config():
    return {
        "authenticator": MagicMock(),
        "start_date": "2021-05-07",
    }


@fixture
def test_current_stream_state():
    return {"updated_time": "2021-10-22"}


@fixture
def test_record():
    return {"items": [{}], "bookmark": "string"}


@fixture
def test_record_filter():
    return {"items": [{"updated_time": "2021-11-01"}], "bookmark": "string"}


@fixture
def test_response(test_record):
    response = MagicMock()
    response.json.return_value = test_record
    return response


@fixture
def test_response_filter(test_record_filter):
    response = MagicMock()
    response.json.return_value = test_record_filter
    return response


@fixture
def analytics_report_stream():
    return CampaignAnalyticsReport(parent=None, config=MagicMock())


@fixture
def date_range():
    return {"start_date": "2023-01-01", "end_date": "2023-01-31", "parent": {"id": "123"}}



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import os
from http import HTTPStatus
from unittest.mock import MagicMock

import pytest
import requests
from source_pinterest.streams import (
    AdAccountAnalytics,
    AdAccounts,
    AdAnalytics,
    AdGroupAnalytics,
    AdGroups,
    Ads,
    Audiences,
    BoardPins,
    Boards,
    BoardSectionPins,
    BoardSections,
    CampaignAnalytics,
    Campaigns,
    Catalogs,
    CatalogsFeeds,
    CatalogsProductGroups,
    ConversionTags,
    CustomerLists,
    Keywords,
    PinterestStream,
    PinterestSubStream,
    RateLimitExceeded,
    UserAccountAnalytics,
)

os.environ["REQUEST_CACHE_PATH"] = "/tmp"


@pytest.fixture
def patch_base_class(mocker):
    # Mock abstract methods to enable instantiating abstract class
    mocker.patch.object(PinterestStream, "path", "v0/example_endpoint")
    mocker.patch.object(PinterestStream, "primary_key", "test_primary_key")
    mocker.patch.object(PinterestStream, "__abstractmethods__", set())
    #
    mocker.patch.object(PinterestSubStream, "path", "v0/example_endpoint")
    mocker.patch.object(PinterestSubStream, "primary_key", "test_primary_key")
    mocker.patch.object(PinterestSubStream, "next_page_token", None)
    mocker.patch.object(PinterestSubStream, "parse_response", {})
    mocker.patch.object(PinterestSubStream, "__abstractmethods__", set())


def test_request_params(patch_base_class):
    stream = PinterestStream(config=MagicMock())
    inputs = {"stream_slice": None, "stream_state": None, "next_page_token": None}
    expected_params = {}
    assert stream.request_params(**inputs) == expected_params


def test_next_page_token(patch_base_class, test_response):
    stream = PinterestStream(config=MagicMock())
    inputs = {"response": test_response}
    expected_token = {"bookmark": "string"}
    assert stream.next_page_token(**inputs) == expected_token


def test_parse_response(patch_base_class, test_response, test_current_stream_state):
    stream = PinterestStream(config=MagicMock())
    inputs = {"response": test_response, "stream_state": test_current_stream_state}
    expected_parsed_object = {}
    assert next(stream.parse_response(**inputs)) == expected_parsed_object


def test_parse_response_with_sensitive_data(patch_base_class):
    """Test that sensitive data is removed"""
    stream = CatalogsFeeds(config=MagicMock())
    response = MagicMock()
    response.json.return_value = {"items": [{"id": "CatalogsFeeds1", "credentials": {"password": "bla"}}], "bookmark": "string"}
    actual_response = list(stream.parse_response(response=response, stream_state=None))
    assert actual_response == [{"id": "CatalogsFeeds1"}]


def test_request_headers(patch_base_class):
    stream = PinterestStream(config=MagicMock())
    inputs = {"stream_slice": None, "stream_state": None, "next_page_token": None}
    expected_headers = {}
    assert stream.request_headers(**inputs) == expected_headers


def test_http_method(patch_base_class):
    stream = PinterestStream(config=MagicMock())
    expected_method = "GET"
    assert stream.http_method == expected_method


@pytest.mark.parametrize(
    ("http_status", "should_retry"),
    [
        (HTTPStatus.OK, False),
        (HTTPStatus.BAD_REQUEST, False),
        (HTTPStatus.TOO_MANY_REQUESTS, False),
        (HTTPStatus.INTERNAL_SERVER_ERROR, True),
    ],
)
def test_should_retry(patch_base_class, http_status, should_retry):
    response_mock = MagicMock()
    response_mock.status_code = http_status
    stream = PinterestStream(config=MagicMock())
    assert stream.should_retry(response_mock) == should_retry


def test_backoff_time(patch_base_class):
    response_mock = MagicMock()
    stream = PinterestStream(config=MagicMock())
    expected_backoff_time = None
    assert stream.backoff_time(response_mock) == expected_backoff_time


@pytest.mark.parametrize(
    "test_response, status_code, expected",
    [
        ({"code": 8, "message": "You have exceeded your rate limit. Try again later."}, 429, False),
        ({"code": 7, "message": "Some other error message"}, 429, False),
    ],
)
def test_should_retry_on_max_rate_limit_error(requests_mock, test_response, status_code, expected):
    stream = Boards(config=MagicMock())
    url = "https://api.pinterest.com/v5/boards"
    requests_mock.get("https://api.pinterest.com/v5/boards", json=test_response, status_code=status_code)
    response = requests.get(url)
    result = stream.should_retry(response)
    assert result == expected


def test_non_json_response(requests_mock):
    stream = UserAccountAnalytics(parent=None, config=MagicMock())
    url = "https://api.pinterest.com/v5/boards"
    requests_mock.get("https://api.pinterest.com/v5/boards", text="some response", status_code=200)
    response = requests.get(url)
    try:
        stream.should_retry(response)
        assert False
    except Exception as e:
        assert "Received unexpected response in non json format" in str(e)


@pytest.mark.parametrize(
    "test_response, test_headers, status_code, expected",
    [
        ({"code": 7, "message": "Some other error message"}, {"X-RateLimit-Reset": "2"}, 429, 2.0),
        (
            {"code": 7, "message": "Some other error message"},
            {"X-RateLimit-Reset": "2000"},
            429,
            (RateLimitExceeded, "Rate limit exceeded for stream boards. Waiting time is longer than 10 minutes: 2000.0s."),
        ),
    ],
)
def test_backoff_on_rate_limit_error(requests_mock, test_response, status_code, test_headers, expected):
    stream = Boards(config=MagicMock())
    url = "https://api.pinterest.com/v5/boards"
    requests_mock.get(
        "https://api.pinterest.com/v5/boards",
        json=test_response,
        headers=test_headers,
        status_code=status_code,
    )

    response = requests.get(url)

    if isinstance(expected, tuple):
        with pytest.raises(expected[0], match=expected[1]):
            stream.backoff_time(response)
    else:
        result = stream.backoff_time(response)
        assert result == expected


@pytest.mark.parametrize(
    ("stream_cls, slice, expected"),
    [
        (Boards(MagicMock()), None, "boards"),
        (AdAccounts(MagicMock()), None, "ad_accounts"),
        (BoardSections(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "boards/123/sections"),
        (BoardPins(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "boards/123/pins"),
        (
            BoardSectionPins(parent=None, config=MagicMock()),
            {"sub_parent": {"parent": {"id": "234"}}, "parent": {"id": "123"}},
            "boards/234/sections/123/pins",
        ),
        (AdAccountAnalytics(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "ad_accounts/123/analytics"),
        (Campaigns(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "ad_accounts/123/campaigns"),
        (
            CampaignAnalytics(parent=None, config=MagicMock()),
            {"sub_parent": {"parent": {"id": "234"}}, "parent": {"id": "123"}},
            "ad_accounts/234/campaigns/analytics",
        ),
        (AdGroups(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "ad_accounts/123/ad_groups"),
        (
            AdGroupAnalytics(parent=None, config=MagicMock()),
            {"sub_parent": {"parent": {"id": "234"}}, "parent": {"id": "123"}},
            "ad_accounts/234/ad_groups/analytics",
        ),
        (Ads(parent=None, config=MagicMock()), {"parent": {"id": "123"}}, "ad_accounts/123/ads"),
        (
            AdAnalytics(parent=None, config=MagicMock()),
            {"sub_parent": {"parent": {"id": "234"}}, "parent": {"id": "123"}},
            "ad_accounts/234/ads/analytics",
        ),
        (Catalogs(config=MagicMock()), None, "catalogs"),
        (CatalogsFeeds(config=MagicMock()), None, "catalogs/feeds"),
        (CatalogsProductGroups(config=MagicMock()), None, "catalogs/product_groups"),
        (
            Keywords(parent=None, config=MagicMock()),
            {"parent": {"id": "234", "ad_account_id": "AD_ACCOUNT_1"}},
            "ad_accounts/AD_ACCOUNT_1/keywords?ad_group_id=234",
        ),
        (Audiences(parent=None, config=MagicMock()), {"parent": {"id": "AD_ACCOUNT_1"}}, "ad_accounts/AD_ACCOUNT_1/audiences"),
        (ConversionTags(parent=None, config=MagicMock()), {"parent": {"id": "AD_ACCOUNT_1"}}, "ad_accounts/AD_ACCOUNT_1/conversion_tags"),
        (CustomerLists(parent=None, config=MagicMock()), {"parent": {"id": "AD_ACCOUNT_1"}}, "ad_accounts/AD_ACCOUNT_1/customer_lists"),
    ],
)
def test_path(patch_base_class, stream_cls, slice, expected):
    stream = stream_cls
    if slice:
        result = stream.path(stream_slice=slice)
    else:
        result = stream.path()
    assert result == expected



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import copy
import os
from unittest.mock import MagicMock

import pytest
import responses
from source_pinterest import SourcePinterest
from source_pinterest.reports import CampaignAnalyticsReport
from source_pinterest.reports.reports import (
    AdGroupReport,
    AdGroupTargetingReport,
    AdvertiserReport,
    AdvertiserTargetingReport,
    CampaignTargetingReport,
    KeywordReport,
    PinPromotionReport,
    PinPromotionTargetingReport,
    ProductGroupReport,
    ProductGroupTargetingReport,
    ProductItemReport,
)
from source_pinterest.utils import get_analytics_columns
from unit_tests.test_source import setup_responses

os.environ["REQUEST_CACHE_PATH"] = '/tmp'

@responses.activate
def test_request_body_json(analytics_report_stream, date_range):
    granularity = "DAY"
    columns = get_analytics_columns()

    expected_body = {
        "start_date": date_range["start_date"],
        "end_date": date_range["end_date"],
        "granularity": granularity,
        "columns": columns.split(","),
        "level": analytics_report_stream.level,
    }

    body = analytics_report_stream.request_body_json(date_range)
    assert body == expected_body


@responses.activate
def test_read_records(analytics_report_stream, date_range):
    report_download_url = "https://download.report"
    report_request_url = "https://api.pinterest.com/v5/ad_accounts/123/reports"

    final_report_status = {"report_status": "FINISHED", "url": report_download_url}

    initial_response = {"report_status": "IN_PROGRESS", "token": "token", "message": ""}

    final_response = {"campaign_id": [{"metric": 1}]}

    responses.add(responses.POST, report_request_url, json=initial_response)
    responses.add(responses.GET, report_request_url, json=final_report_status, status=200)
    responses.add(responses.GET, report_download_url, json=final_response, status=200)

    sync_mode = "full_refresh"
    cursor_field = ["last_updated"]
    stream_state = {
        "start_date": "2023-01-01",
        "end_date": "2023-01-31",
    }

    records = analytics_report_stream.read_records(sync_mode, cursor_field, date_range, stream_state)
    expected_record = {"metric": 1}

    assert next(records) == expected_record
    assert len(responses.calls) == 3
    assert responses.calls[0].request.url == report_request_url


@responses.activate
def test_streams(test_config):
    setup_responses()
    source = SourcePinterest()
    streams = source.streams(test_config)
    expected_streams_number = 32
    assert len(streams) == expected_streams_number

@responses.activate
def test_custom_streams(test_config):
    config = copy.deepcopy(test_config)
    config['custom_reports'] = [{
        "name": "vadim_report",
        "level": "AD_GROUP",
        "granularity": "MONTH",
        "click_window_days": 30,
        "engagement_window_days": 30,
        "view_window_days": 30,
        "conversion_report_time": "TIME_OF_CONVERSION",
        "attribution_types": ["INDIVIDUAL", "HOUSEHOLD"],
        "columns": ["ADVERTISER_ID", "AD_ACCOUNT_ID", "AD_GROUP_ID", "CTR", "IMPRESSION_2"],
        "start_date": "2023-01-08"
    }]
    setup_responses()
    source = SourcePinterest()
    streams = source.streams(config)
    expected_streams_number = 33
    assert len(streams) == expected_streams_number

@pytest.mark.parametrize(
    "report_name, expected_level",
    [
        [CampaignAnalyticsReport, 'CAMPAIGN'],
        [CampaignTargetingReport, 'CAMPAIGN_TARGETING'],
        [AdvertiserReport, 'ADVERTISER'],
        [AdvertiserTargetingReport, 'ADVERTISER_TARGETING'],
        [AdGroupReport, 'AD_GROUP'],
        [AdGroupTargetingReport, 'AD_GROUP_TARGETING'],
        [PinPromotionReport, 'PIN_PROMOTION'],
        [PinPromotionTargetingReport, 'PIN_PROMOTION_TARGETING'],
        [ProductGroupReport, 'PRODUCT_GROUP'],
        [ProductGroupTargetingReport, 'PRODUCT_GROUP_TARGETING'],
        [ProductItemReport, 'PRODUCT_ITEM'],
        [KeywordReport, 'KEYWORD']
    ],
)
def test_level(test_config, report_name, expected_level):
    assert report_name(parent=None, config=MagicMock()).level == expected_level

